#include <Wire.h> //Used to comuicate with the MPU 6050 using I2C

byte last_channel_1, last_channel_2, last_channel_3, last_channel_4;
volatile int receiver_channels[4];
unsigned long timer1, timer2, timer3, timer4, current_time, motor_time, timer_channel_1, timer_channel_2, timer_channel_3, timer_channel_4;
int esc1, esc2, esc3, esc4;
unsigned long Est = 10000000;
unsigned long before_init, refresh_timer, esc_loop_timer;
int deadband = 8;
bool arm = 0;

//Delcaering the varuables used for the self leveling

float PID_pitch;
float pid_kp_pitch = 3.5;     //The kp, ki, and kp values (the inital valuses of my mpu-6050 due to unlevel surface)
float pid_ki_pitch = 0.001;
float pid_kd_pitch = 2.00;
float pid_i_pitch = 0;
float pitch_setpoint = 0;
float pitch_error;
float pitch_prev_error = 0;
float pitch_angle;

float PID_roll;
float pid_kp_roll = 3.5;
float pid_ki_roll = 0.001;
float pid_kd_roll = 2.00;
float pid_i_roll = 0;
float roll_setpoint = 0;
float roll_error;
float roll_prev_error = 0;
float roll_angle;

float PID_yaw;
float pid_kp_yaw = 2.00;
float pid_ki_yaw = 0.00;
float pid_kd_yaw = 1.00;
float pid_i_yaw = 0;
float yaw_setpoint = 0;
float yaw_error;
float yaw_prev_error = 0;
float yaw_rate;

int throttle;

int16_t Acc_rawX, Acc_rawY, Acc_rawZ, Gyr_rawX, Gyr_rawY, Gyr_rawZ, temp;
float Acc_X, Acc_Y, Gyr_X, Gyr_Y, Gyr_Z;

void setup() {
  DDRD |= B11110000; //Declare motor pins as outputs
  Wire.begin(); //Initiate I2C with the gyro

  TWBR = 12; //this sets up 400hz compuincation with the gyro

  initGyro();

  //Attachs interrupts
  PCICR |= (1 << PCIE0); 
  PCMSK0 |= (1 << PCINT1);
  PCMSK0 |= (1 << PCINT2);
  PCMSK0 |= (1 << PCINT3);
  PCMSK0 |= (1 << PCINT4);

  refresh_timer = micros();
}

void loop() {
  //To make sure the drone doesn't have false starts
  if (arm == 0) {
    esc1 = 1000;
    esc2 = 1000;
    esc3 = 1000;
    esc4 = 1000;
    if (receiver_channels[0] <= 1050 && receiver_channels[1] <= 1050 && receiver_channels[2] <= 1050 && receiver_channels[3] >= 1950) arm = 1;
  }
  
    Acc_X = atan((Acc_rawY / 4096.0) / sqrt(pow((Acc_rawX / 4096.0), 2) + pow((Acc_rawZ / 4096.0), 2))) * 57.296;
    Acc_Y = atan((Acc_rawX / 4096.0) / sqrt(pow((Acc_rawY / 4096.0), 2) + pow((Acc_rawZ / 4096.0), 2))) * -57.296;
    pitch_angle += Gyr_rawX * 0.0000611;
    roll_angle += Gyr_rawY * 0.0000611; //0.0000611 = 0.004 / 65.5
    yaw_rate = Gyr_rawZ * 0.0153; // 0.0153 = 1 / 65.5

    //Complementary Filter to filter out the correct angle
    pitch_angle = (0.98 * pitch_angle) + (0.02 * Acc_X);
    roll_angle = (0.98 * roll_angle) + (0.02 * Acc_Y);

    //Calculate PIDs
    PID();

    esc1 = throttle - PID_pitch - PID_roll - PID_yaw; //Add the PIDs
    esc2 = throttle - PID_pitch + PID_roll + PID_yaw;
    esc3 = throttle + PID_pitch + PID_roll - PID_yaw;
    esc4 = throttle + PID_pitch - PID_roll + PID_yaw;

    if (esc1 < 1100) esc1 = 1100;                                    //Set limits
    else if (esc1 > 2000) esc1 = 2000;
    if (esc2 < 1100) esc2 = 1100;
    else if (esc2 > 2000) esc2 = 2000;
    if (esc3 < 1100) esc3 = 1100;
    else if (esc3 > 2000) esc3 = 2000;
    if (esc4 < 1100) esc4 = 1100;
    else if (esc4 > 2000) esc4 = 2000;
    if (receiver_channels[0] >= 1950 && receiver_channels[1] <= 1050 && receiver_channels[2] <= 1050 && receiver_channels[3] <= 1050) arm = 0;
  }
  if ((micros() - refresh_timer) > 4050) digitalWrite(13, HIGH);
  while ((micros() - refresh_timer) < 4000);                      //Wait for 4000 us to pass
  refresh_timer = micros();                                       //Reset timer

  PORTD |= B11110000;                                             //Turn on pulse
  timer_channel_1 = esc1 + refresh_timer;                         //Add throttle pulse time to loop time
  timer_channel_2 = esc2 + refresh_timer;
  timer_channel_3 = esc3 + refresh_timer;
  timer_channel_4 = esc4 + refresh_timer;
  while (PORTD >= 16) {
    esc_loop_timer = micros();
    if (timer_channel_1 <= esc_loop_timer)PORTD &= B11101111;     //When loop times equal throttle times, turn off pulse. This essentially creates the 1000-2000 microsecond pulse
    if (timer_channel_2 <= esc_loop_timer)PORTD &= B11011111;
    if (timer_channel_3 <= esc_loop_timer)PORTD &= B10111111;
    if (timer_channel_4 <= esc_loop_timer)PORTD &= B01111111;
  }
}
ISR(PCINT0_vect) {
  //reads all the individual channels
  current_time = micros() % Est;
  //Channel 1
  if (last_channel_1 == 0 && PINB & B00010000) {
    last_channel_1 = 1;
    timer1 = current_time;
  }
  else if (last_channel_1 == 1 && !(PINB & B00010000)) {
    last_channel_1 = 0;
    receiver_channels[0] = ((current_time - timer1) > 0) ? (current_time - timer1) : (Est - timer1 + current_time); //Prevents timer from overflowing
  }
  //Channel 2
  if (last_channel_2 == 0 && PINB & B00001000) {
    last_channel_2 = 1;
    timer2 = current_time;
  }
  else if (last_channel_2 == 1 && !(PINB & B00001000)) {
    last_channel_2 = 0;
    receiver_channels[1] = ((current_time - timer2) > 0) ? (current_time - timer2) : (Est - timer2 + current_time);
  }
  //Channel 3
  if (last_channel_3 == 0 && PINB & B00000100) {
    last_channel_3 = 1;
    timer3 = current_time;
  }
  else if (last_channel_3 == 1 && !(PINB & B00000100)) {
    last_channel_3 = 0;
    receiver_channels[2] = ((current_time - timer3) > 0) ? (current_time - timer3) : (Est - timer3 + current_time);
  }
  //Channel 4
  if (last_channel_4 == 0 && PINB & B00000010) {
    last_channel_4 = 1;
    timer4 = current_time;
  }
  else if (last_channel_4 == 1 && !(PINB & B00000010)) {
    last_channel_4 = 0;
    receiver_channels[3] = ((current_time - timer4) > 0) ? (current_time - timer4) : (Est - timer4 + current_time);
  }
}

//reading the values from the MPU 6050
void readGyro() {
  Wire.beginTransmission(0x68);
  Wire.write(0x3B);
  Wire.endTransmission(false);
  Wire.requestFrom(0x68, 14, true);
  Acc_rawX = Wire.read() << 8 | Wire.read();                    //Get Most significant bits, Shift over, and add least significant bits
  Acc_rawY = Wire.read() << 8 | Wire.read();
  Acc_rawZ = Wire.read() << 8 | Wire.read();
  temp = Wire.read() << 8 | Wire.read();
  Gyr_rawX = Wire.read() << 8 | Wire.read();                    //Get Most significant bits, Shift over, and add least significant bits
  Gyr_rawY = Wire.read() << 8 | Wire.read();
  Gyr_rawZ = Wire.read() << 8 | Wire.read();

  //clamps the range of RC receiver signals within 1000-2000 us
  for (int a = 0; a < 4; a++) {
    if (receiver_channels[a] > 2000) {
      receiver_channels[a] = 2000;
    }
    else if (receiver_channels[a] < 1000) {
      receiver_channels[a] = 1000;
    }
    else if ((receiver_channels[a] <= (1500 + deadband)) && (receiver_channels[a] >= (1500 - deadband))) {
      receiver_channels[a] = 1500;
    }
  }
}

//I got the adresses from the MPU 6050 datasheet
void initGyro() {
  Wire.beginTransmission(0x68);             //Start communicating with the IMU and set parameters (This can differ)
  Wire.write(0x37);
  Wire.write(0x02);
  Wire.endTransmission();
  Wire.beginTransmission(0x68);
  Wire.write(0x6A);
  Wire.write(0x00);
  Wire.endTransmission();
  Wire.beginTransmission(0x68);
  Wire.write(0x6B);
  Wire.write(0x00);
  Wire.endTransmission();
  Wire.beginTransmission(0x68);
  Wire.write(0x1B);
  Wire.write(0x08);
  Wire.endTransmission();
  Wire.beginTransmission(0x68);
  Wire.write(0x1C);
  Wire.write(0x10);
  Wire.endTransmission();
  Wire.beginTransmission(0x68);
  Wire.write(0x1A);
  Wire.write(0x03);
  Wire.endTransmission();

  Wire.beginTransmission(0x1E);
  Wire.write(0x00);
  Wire.write(0x78);
  Wire.endTransmission();
  Wire.beginTransmission(0x1E);
  Wire.write(0x20);
  Wire.write(0x00);
  Wire.endTransmission();
}

//The self leveling calculations
void PID(){
  pitch_error = pitch_angle - pitch_setpoint;
  pid_i_pitch += pid_ki_pitch * pitch_error;
  PID_pitch = (pid_kp_pitch * pitch_error) + pid_i_pitch + (pid_kd_pitch * (pitch_error - pitch_prev_error));
  pitch_prev_error = pitch_error;
  if (PID_pitch < -400)PID_pitch = -400;
  else if (PID_pitch > 400)PID_pitch = 400;

  roll_error = roll_angle - roll_setpoint;
  pid_i_roll += pid_ki_roll * roll_error;
  PID_roll = (pid_kp_roll * roll_error) + pid_i_roll + (pid_kd_roll * (roll_error - roll_prev_error));
  roll_prev_error = roll_error;
  if (PID_roll < -400)PID_roll = -400;
  else if (PID_roll > 400)PID_roll = 400;

  yaw_error = yaw_rate - yaw_setpoint;
  pid_i_yaw += pid_ki_yaw * yaw_error;
  PID_yaw = (pid_kp_yaw * yaw_error) + pid_i_yaw + (pid_kd_yaw * (yaw_error - yaw_prev_error));
  yaw_prev_error = yaw_error;
  if (PID_yaw < -400)PID_yaw = -400;
  else if (PID_yaw > 400)PID_yaw = 400;
}
